module.exports = function ({types}) {
  return {
    visitor: {
      Function: function (path) {
        let functionName = '';

        if (path.node.id) {
          functionName = path.node.id.name;
        } else if (path.node.key) {
          functionName = path.node.key.name;
        }

        (path.get('params') || [])
          .slice()
          .forEach(function (param) {
            const name = param.node.name;

            let resultantDecorator;

            (param.node.decorators || [])
              .slice()
              .forEach(function (decorator) {
                const classDeclaration = path.findParent(p => p.node.type === 'VariableDeclaration');
                const classDeclarator = path.findParent(p => p.node.type === 'VariableDeclarator');
                const className = classDeclarator.node.id.name;

                if (functionName === className) {
                  resultantDecorator = types.callExpression(
                    decorator.expression, [
                      types.Identifier(className),
                      types.Identifier('undefined'),
                      types.NumericLiteral(param.key)
                    ]
                  );
                } else {
                  resultantDecorator = types.callExpression(
                    decorator.expression, [
                      types.Identifier(`${className}.prototype`),
                      types.StringLiteral(functionName),
                      types.NumericLiteral(param.key)
                    ]
                  );
                }

                const expression = types.expressionStatement(resultantDecorator);

                classDeclaration.insertAfter(expression);
              });

            if (resultantDecorator) {
              param.replaceWith(types.Identifier(name));
            }
          });

      }
    }
  }
};
